"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.evaluateFixtures = evaluateFixtures;

var _path = require("path");

var _fsExtra = require("fs-extra");

var _graphql = require("graphql");

var _graphqlConfig = require("graphql-config");

var _graphqlConfigUtilities = require("graphql-config-utilities");

var _graphqlToolUtilities = require("graphql-tool-utilities");

var _validate = require("./validate.js");

async function evaluateFixtures(fixturePaths, {
  cwd
}) {
  const config = (0, _graphqlConfig.getGraphQLConfig)((0, _path.resolve)(cwd));
  const projectASTCollection = await Promise.all((0, _graphqlConfigUtilities.getGraphQLProjects)(config).map(getOperationsForProject));
  return runForEachFixture(fixturePaths, fixture => evaluateFixture(fixture, projectASTCollection));
}

async function getOperationsForProject(projectConfig) {
  const operationPaths = await (0, _graphqlConfigUtilities.getGraphQLProjectIncludedFilePaths)(projectConfig);
  const operationSources = await Promise.all(operationPaths.map(loadOperationSource));
  const document = (0, _graphql.concatAST)(operationSources.map(source => {
    try {
      return (0, _graphql.parse)(source);
    } catch (error) {
      throw new Error(`Error parsing '${source.name}':\n\n${error.message.replace(/Syntax Error.*?\(.*?\) /, '')}`);
    }
  }));
  let schema;

  try {
    schema = projectConfig.getSchema();
  } catch (error) {
    throw new Error(`Error parsing '${projectConfig.schemaPath}':\n\n${error.message.replace(/Syntax Error.*?\(.*?\) /, '')}`);
  }

  return {
    ast: (0, _graphqlToolUtilities.compile)(schema, document),
    config: projectConfig
  };
}

async function loadOperationSource(filePath) {
  const body = await (0, _fsExtra.readFile)(filePath, 'utf8');
  return new _graphql.Source(body, filePath);
}

function runForEachFixture(fixturePaths, runner) {
  return Promise.all(fixturePaths.map(async fixturePath => {
    try {
      const fixture = await (0, _fsExtra.readJSON)(fixturePath);
      return {
        fixturePath,
        ...runner({
          path: fixturePath,
          content: fixture
        })
      };
    } catch (error) {
      return {
        fixturePath,
        scriptError: error,
        validationErrors: []
      };
    }
  }));
}

function evaluateFixture(fixture, projectASTCollection) {
  try {
    const {
      operation,
      projectAST: {
        ast
      }
    } = (0, _validate.getOperationForFixture)(fixture, projectASTCollection);
    return (0, _validate.validateFixture)(fixture, ast, operation);
  } catch (error) {
    if (error instanceof _validate.MissingOperationError || error instanceof _validate.AmbiguousOperationNameError) {
      return {
        fixturePath: fixture.path,
        validationErrors: [error]
      };
    }

    throw error;
  }
}