"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateFixture = exports.getOperationForFixture = exports.findOperations = exports.getOperationNames = exports.AmbiguousOperationNameError = exports.MissingOperationError = void 0;
var tslib_1 = require("tslib");
var path_1 = require("path");
var graphql_1 = require("graphql");
var graphql_config_utilities_1 = require("graphql-config-utilities");
var MissingOperationError = /** @class */ (function (_super) {
    tslib_1.__extends(MissingOperationError, _super);
    function MissingOperationError(_a, lookForOperationNames, projectASTCollection) {
        var path = _a.path;
        return _super.call(this, [
            "Could not find a matching operation for '" + path + "'",
            "(looked for " + lookForOperationNames.join(', ') + ").",
            "Make sure to put your fixture in a folder named the same as the operation,",
            "or add an '" + OPERATION_MARKER + "' key indicating the operation.",
            "Available operations: " + projectASTCollection
                .flatMap(function (_a) {
                var operations = _a.ast.operations;
                return Object.keys(operations);
            })
                .join(', '),
        ].join(' ')) || this;
    }
    return MissingOperationError;
}(Error));
exports.MissingOperationError = MissingOperationError;
var AmbiguousOperationNameError = /** @class */ (function (_super) {
    tslib_1.__extends(AmbiguousOperationNameError, _super);
    function AmbiguousOperationNameError(_a, foundOperations) {
        var path = _a.path;
        return _super.call(this, [
            "Ambiguous operation name found for '" + path + "'",
            "(found " + Array.from(new Set(foundOperations.map(function (_a) {
                var lookForOperationName = _a.lookForOperationName;
                return lookForOperationName;
            }))).join(', ') + ")",
            "in projects:",
            foundOperations
                .map(function (_a) {
                var config = _a.projectAST.config;
                return graphql_config_utilities_1.resolveProjectName(config);
            })
                .join(', ') + ".",
            "Try renaming the operation in one of the projects listed and updating",
            "the fixture folder name or use an '" + OPERATION_MARKER + "' key indicating",
            "the new operation name.",
        ].join(' ')) || this;
    }
    return AmbiguousOperationNameError;
}(Error));
exports.AmbiguousOperationNameError = AmbiguousOperationNameError;
var OPERATION_MARKER = '@operation';
function normalizeOperationName(operationName) {
    return operationName
        ? operationName.replace(/(Query|Mutation|Subscription)$/i, '')
        : undefined;
}
function getOperationNames(fixture) {
    var fixtureDirectoryName = path_1.basename(path_1.dirname(fixture.path));
    var operationMarkerName = fixture.content[OPERATION_MARKER];
    return Array.from(new Set([
        fixtureDirectoryName,
        operationMarkerName,
        normalizeOperationName(fixtureDirectoryName),
        normalizeOperationName(operationMarkerName),
    ])).filter(function (operationName) {
        return typeof operationName === 'string';
    });
}
exports.getOperationNames = getOperationNames;
function findOperations(lookForOperationNames, projectASTCollection) {
    return projectASTCollection
        .map(function (projectAST) {
        var e_1, _a;
        try {
            for (var lookForOperationNames_1 = tslib_1.__values(lookForOperationNames), lookForOperationNames_1_1 = lookForOperationNames_1.next(); !lookForOperationNames_1_1.done; lookForOperationNames_1_1 = lookForOperationNames_1.next()) {
                var lookForOperationName = lookForOperationNames_1_1.value;
                var operation = projectAST.ast.operations[lookForOperationName];
                if (operation) {
                    return {
                        lookForOperationName: lookForOperationName,
                        operation: operation,
                        projectAST: projectAST,
                    };
                }
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (lookForOperationNames_1_1 && !lookForOperationNames_1_1.done && (_a = lookForOperationNames_1.return)) _a.call(lookForOperationNames_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return null;
    })
        .filter(function (match) { return Boolean(match); });
}
exports.findOperations = findOperations;
function getOperationForFixture(fixture, projectASTCollection) {
    var lookForOperationNames = getOperationNames(fixture);
    var operations = findOperations(lookForOperationNames, projectASTCollection);
    if (operations.length === 0) {
        throw new MissingOperationError(fixture, lookForOperationNames, projectASTCollection);
    }
    if (operations.length > 1) {
        throw new AmbiguousOperationNameError(fixture, operations);
    }
    return operations[0];
}
exports.getOperationForFixture = getOperationForFixture;
function validateFixture(fixture, ast, operation) {
    var _a = operation.fields, fields = _a === void 0 ? [] : _a, filePath = operation.filePath, operationType = operation.operationType;
    var value = tslib_1.__assign({}, fixture.content);
    delete value[OPERATION_MARKER];
    return {
        fixturePath: fixture.path,
        operationName: operation.operationName,
        operationType: operationType,
        operationPath: filePath === 'GraphQL request' ? undefined : filePath,
        validationErrors: fields.reduce(function (allErrors, field) {
            return allErrors.concat(validateValueAgainstFieldDescription(value[field.responseName], field, '', ast));
        }, []),
    };
}
exports.validateFixture = validateFixture;
function validateValueAgainstFieldDescription(value, fieldDescription, parentKeyPath, ast) {
    var type = fieldDescription.type, responseName = fieldDescription.responseName;
    var keyPath = updateKeyPath(parentKeyPath, responseName);
    var typeErrors = validateValueAgainstType(value, type, keyPath, {
        shallow: true,
    });
    if (typeErrors.length > 0) {
        return typeErrors;
    }
    return Array.isArray(value)
        ? validateListAgainstFieldDescription(value, fieldDescription, type, keyPath, ast)
        : validateValueAgainstObjectFieldDescription(value, fieldDescription, keyPath, ast);
}
function validateValueAgainstObjectFieldDescription(value, fieldDescription, keyPath, ast) {
    if (value == null || typeof value !== 'object' || Array.isArray(value)) {
        return [];
    }
    var _a = fieldDescription.fields, fields = _a === void 0 ? [] : _a, _b = fieldDescription.fragmentSpreads, fragmentSpreads = _b === void 0 ? [] : _b, type = fieldDescription.type;
    var fragmentFields = [];
    if (fragmentSpreads) {
        fragmentSpreads
            .map(function (spread) { return ast.fragments[spread]; })
            .forEach(function (fragment) {
            fragment.fields.forEach(function (field) {
                if (fields.some(function (_a) {
                    var fieldName = _a.fieldName;
                    return fieldName === field.fieldName;
                })) {
                    return;
                }
                var isGuaranteedTypeMatch = fragment.possibleTypes.includes(makeTypeNullable(type));
                fragmentFields.push(isGuaranteedTypeMatch
                    ? field
                    : tslib_1.__assign(tslib_1.__assign({}, field), { type: makeTypeNullable(field.type) }));
            });
        });
    }
    return validateValueAgainstFields(value, fields.concat(fragmentFields), keyPath, ast);
}
function makeTypeNullable(type) {
    return graphql_1.isNonNullType(type) ? type.ofType : type;
}
function validateListAgainstFieldDescription(value, fieldDescription, type, keyPath, ast) {
    var itemType = graphql_1.isNonNullType(type)
        ? type.ofType.ofType
        : type.ofType;
    return value.reduce(function (allErrors, item, index) {
        var itemKeyPath = updateKeyPath(keyPath, index);
        var itemTypeErrors = validateValueAgainstType(item, itemType, itemKeyPath, { shallow: true });
        if (itemTypeErrors.length > 0) {
            return allErrors.concat(itemTypeErrors);
        }
        return Array.isArray(item)
            ? allErrors.concat(validateListAgainstFieldDescription(item, fieldDescription, itemType, itemKeyPath, ast))
            : allErrors.concat(validateValueAgainstObjectFieldDescription(item, fieldDescription, itemKeyPath, ast));
    }, []);
}
function validateValueAgainstFields(value, fields, keyPath, ast) {
    var finalValue = value || {};
    var excessFields = Object.keys(finalValue)
        .filter(function (key) { return fields.every(function (_a) {
        var responseName = _a.responseName;
        return key !== responseName;
    }); })
        .map(function (key) { return error(keyPath, "has key '" + key + "' not present in the query"); });
    return fields.reduce(function (allErrors, field) {
        return allErrors.concat(validateValueAgainstFieldDescription(finalValue[field.responseName], field, keyPath, ast));
    }, excessFields);
}
function validateValueAgainstType(value, type, keyPath, options) {
    if (options === void 0) { options = { shallow: false }; }
    var shallow = options.shallow;
    if (graphql_1.isNonNullType(type)) {
        return value == null
            ? [error(keyPath, "should be non-null but was " + String(value))]
            : validateValueAgainstType(value, type.ofType, keyPath, options);
    }
    if (value === null) {
        return [];
    }
    var valueType = typeof value;
    if (graphql_1.isListType(type)) {
        if (!Array.isArray(value)) {
            return [
                error(keyPath, "should be an array (or null), but was " + articleForType(valueType) + " " + valueType),
            ];
        }
        return shallow
            ? []
            : value.reduce(function (allErrors, item, index) {
                return allErrors.concat(validateValueAgainstType(item, type.ofType, updateKeyPath(keyPath, index), options));
            }, []);
    }
    if (value === undefined) {
        var typeName = nameForType(type);
        return [
            error(keyPath, "should be " + articleForType(typeName) + " " + typeName + " (or null), but was undefined"),
        ];
    }
    if (graphql_1.isObjectType(type)) {
        if (valueType === 'object') {
            if (shallow) {
                return [];
            }
            else {
                var fields_1 = type.getFields();
                return Object.keys(value).reduce(function (fieldErrors, key) {
                    var fieldKeyPath = updateKeyPath(keyPath, key);
                    return fields_1[key] == null
                        ? fieldErrors.concat([
                            error(fieldKeyPath, "does not exist on type " + type.name + " (available fields: " + Object.keys(fields_1).join(', ') + ")"),
                        ])
                        : fieldErrors.concat(validateValueAgainstType(value[key], fields_1[key].type, fieldKeyPath, options));
                }, []);
            }
        }
        else {
            return [error(keyPath, "should be an object but was a " + valueType)];
        }
    }
    if (type === graphql_1.GraphQLString) {
        return valueType === 'string'
            ? []
            : [error(keyPath, "should be a string but was a " + valueType)];
    }
    else if (type === graphql_1.GraphQLInt) {
        if (typeof value === 'number') {
            return Number.isInteger(value)
                ? []
                : [error(keyPath, 'should be an integer but was a float')];
        }
        return [error(keyPath, "should be an integer but was a " + valueType)];
    }
    else if (type === graphql_1.GraphQLFloat) {
        return Number.isNaN(value)
            ? [error(keyPath, "should be a float but was a " + valueType)]
            : [];
    }
    else if (type === graphql_1.GraphQLBoolean) {
        return typeof value === 'boolean'
            ? []
            : [error(keyPath, "should be a boolean but was a " + valueType)];
    }
    if (graphql_1.isScalarType(type)) {
        return type.parseValue(value) == null
            ? [error(keyPath, "value does not match scalar " + nameForType(type))]
            : [];
    }
    if (graphql_1.isEnumType(type)) {
        return type.parseValue(value) == null
            ? [
                error(keyPath, "value does not match enum " + nameForType(type) + " (available values: " + type
                    .getValues()
                    .map(function (enumValue) { return enumValue.value; })
                    .join(', ') + ")"),
            ]
            : [];
    }
    return [];
}
var CUSTOM_NAMES = (_a = {},
    _a[graphql_1.GraphQLBoolean.name] = 'boolean',
    _a[graphql_1.GraphQLFloat.name] = 'float',
    _a[graphql_1.GraphQLInt.name] = 'integer',
    _a[graphql_1.GraphQLString.name] = 'string',
    _a);
function nameForType(type) {
    return Object.prototype.hasOwnProperty.call(CUSTOM_NAMES, type.name)
        ? CUSTOM_NAMES[type.name]
        : type.name;
}
var TYPES_WITH_ARTICLE_AN = ['object', 'integer', 'array'];
function articleForType(type) {
    return TYPES_WITH_ARTICLE_AN.includes(type) ? 'an' : 'a';
}
function updateKeyPath(keyPath, newKey) {
    if (typeof newKey === 'number') {
        return keyPath + "[" + newKey + "]";
    }
    return keyPath ? keyPath + "." + newKey : newKey;
}
function error(keyPath, message) {
    return { keyPath: keyPath, message: message };
}
